@@ -8,17 +8,15 @@
  */
 #include "config.h"
 
-#include <string.h>
-
-#include <libgimp/gimp.h>
-#include <libgimp/gimpui.h>
-
 #include "main.h"
 #include "interface.h"
 #include "render.h"
-
 #include "plugin-intl.h"
 
+#include <string.h>
+#include <libgimp/gimp.h>
+#include <libgimp/gimpui.h>
+
 
 /*  Constants  */
 
@@ -39,9 +37,9 @@ static void run(const gchar *name, gint nparams, const GimpParam *param, gint *n
 /*  Local variables  */
 
 const PlugInVals default_vals = {
-	-1,  /* chanindex */
-	FALSE,  /* usesel */
-	FALSE,  /* halfsel */
+	-1,  /* chanid */
+	TRUE,  /* usesel */
+	TRUE,  /* halfsel */
 
 	64,  /* window */
 	3.0,  /* sigma */
@@ -51,21 +49,21 @@ const PlugInVals default_vals = {
 	TRUE,  /* slopes */
 	TRUE,  /* leastsq */
 	2,  /* linwindow */
-	//TRUE,  /* lincenter */
-	FALSE,  /* linsymwin */
 	3.0,  /* linsigma */
 	KERNEL_SQUARE,  /* linkernel */
 	FALSE,  /* linnewlayer */
 
 	FALSE,  /* changesel */
-	0,  /* lowthres */
+	0.0,  /* lowthres */
 	0.01,  /* highthres */
 	FALSE,  /* logthres */
 	FALSE,  /* multmask */
+
+	FALSE,  /* all_white_input */
 };
 
 const PlugInUIVals default_ui_vals = {
-	FALSE,
+	TRUE,  /* preview */
 };
 
 static PlugInVals vals;
@@ -86,15 +84,31 @@ static void query(void) {
 	gchar *help_uri;
 
 	static GimpParamDef args[] = {
-		{GIMP_PDB_INT32, "run_mode",
-		 "Interactive, non-interactive"},
-		{GIMP_PDB_IMAGE, "image", "Input image"},
-		{GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
-		/*{ GIMP_PDB_INT32,    "dummy",      "dummy1"                          },
-		   { GIMP_PDB_INT32,    "dummy",      "dummy2"                          },
-		   { GIMP_PDB_INT32,    "dummy",      "dummy3"                          },
-		   { GIMP_PDB_INT32,    "seed",       "Seed value (used only if randomize is FALSE)" },
-		   { GIMP_PDB_INT32,    "randomize",  "Use a random seed (TRUE, FALSE)" } */
+		{ GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
+		{ GIMP_PDB_IMAGE, "image", "Input image" },
+		{ GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
+		{ GIMP_PDB_CHANNEL, "chanid", "Input mask" },
+		{ GIMP_PDB_INT8, "usesel", "Subtract selection from input mask" },
+		{ GIMP_PDB_INT8, "halfsel", "Read pixels with input mask<255" },
+		{ GIMP_PDB_INT32, "window", "Size of window, around each pixel, for taking weighted average" },
+		{ GIMP_PDB_FLOAT, "sigma", "Parameter for kernel" },
+		{ GIMP_PDB_INT32, "kernel", "Type of kernel" },
+		{ GIMP_PDB_INT8, "newlayer", "New layer for interpolation result" },
+
+		{ GIMP_PDB_INT8, "slopes", "Do slopes calculation" },
+		{ GIMP_PDB_INT8, "leastsq", "Least squares estimation for slopes" },
+		{ GIMP_PDB_INT32, "linwindow", "Window for slopes estimation" },
+		{ GIMP_PDB_FLOAT, "linsigma", "Parameter for kernel for slopes estimation" },
+		{ GIMP_PDB_INT32, "linkernel", "Type of kernel for slopes estimation" },
+		{ GIMP_PDB_INT8, "linnewlayer", "Create new grain merge layer for slopes" },
+
+		{ GIMP_PDB_INT8, "changesel", "Modify selection" },
+		{ GIMP_PDB_FLOAT, "lowthres", "Weight sum that gives completely selected pixels" },
+		{ GIMP_PDB_FLOAT, "highthres", "Weight sum that gives completely unselected pixels" },
+		{ GIMP_PDB_INT8, "logthres", "Log scale new selection" },
+		{ GIMP_PDB_INT8, "multmask", "Multiply new selection with mask" },
+
+		{ GIMP_PDB_INT8, "all_white_input", "Use all-white input mask, instead of selected channel" },
 	};
 
 	gimp_plugin_domain_register(PLUGIN_NAME, LOCALEDIR);
@@ -103,8 +117,9 @@ static void query(void) {
 	help_uri = g_filename_to_uri(help_path, NULL, NULL);
 	g_free(help_path);
 
-	//gimp_plugin_help_register
-	//    ("http://developer.gimp.org/plug-in-template/help", help_uri);
+	/* help is in tooltips */
+	/*gimp_plugin_help_register
+	    ("http://developer.gimp.org/plug-in-template/help", help_uri);*/
 
 	gimp_install_procedure(PROCEDURE_NAME,
 			       "Simple interpolatin",
@@ -158,8 +173,30 @@ static void run(const gchar *name,
 		if (n_params != 8) {
 			status = GIMP_PDB_CALLING_ERROR;
 		} else {
-			//if (vals.random_seed)
-			//vals.seed = g_random_int ();
+			vals.chanid = param[3].data.d_channel;
+			vals.usesel = param[4].data.d_int8;
+			vals.halfsel = param[5].data.d_int8;
+			vals.window = param[6].data.d_int32;
+			vals.sigma = param[7].data.d_float;
+			vals.kernel = param[8].data.d_int32;
+			vals.newlayer = param[9].data.d_int8;
+
+			vals.slopes = param[10].data.d_int8;
+			vals.leastsq = param[11].data.d_int8;
+			vals.linwindow = param[12].data.d_int32;
+			vals.linsigma = param[13].data.d_float;
+			vals.linkernel = param[14].data.d_int32;
+			vals.linnewlayer = param[15].data.d_int8;
+
+			vals.changesel = param[16].data.d_int8;
+			vals.lowthres = param[17].data.d_float;
+			vals.highthres = param[18].data.d_float;
+			vals.logthres = param[19].data.d_int8;
+			vals.multmask = param[20].data.d_int8;
+
+			vals.all_white_input = param[21].data.d_int8;
+
+			/*if (vals.random_seed) vals.seed = g_random_int ();*/
 		}
 		break;
 
@@ -177,8 +214,7 @@ static void run(const gchar *name,
 		/*  Possibly retrieve data  */
 		gimp_get_data(DATA_KEY_VALS, &vals);
 
-		//if (vals.random_seed)
-		//vals.seed = g_random_int ();
+		/*if (vals.random_seed) vals.seed = g_random_int ();*/
 		break;
 
 	default:
@@ -186,7 +222,7 @@ static void run(const gchar *name,
 	}
 
 	if (status == GIMP_PDB_SUCCESS) {
-		render(image_ID, drawable, &vals);
+		render(drawable, &vals, 0);
 
 		if (run_mode != GIMP_RUN_NONINTERACTIVE)
 			gimp_displays_flush();
