@@ -9,6 +9,9 @@
 #ifndef __MAIN_H__
 #define __MAIN_H__
 
+#include <libgimp/gimp.h>
+
+
 typedef enum InterpolateKernelType {
 	KERNEL_CONSTANT,
 	KERNEL_LINEAR,
@@ -18,6 +21,7 @@ typedef enum InterpolateKernelType {
 	KERNEL_EXPONENTIAL,
 	KERNEL_GAUSSIAN,
 	KERNEL_INVNTH,
+	KERNEL_BRUSH,
 } InterpolateKernelType;
 
 typedef struct {
@@ -33,7 +37,6 @@ typedef struct {
 	gboolean slopes;
 	gboolean leastsq;
 	gint linwindow;
-	gboolean linsymwin;
 	gfloat linsigma;
 	InterpolateKernelType linkernel;
 	gboolean linnewlayer;
@@ -43,10 +46,12 @@ typedef struct {
 	gfloat highthres;
 	gboolean logthres;
 	gboolean multmask;
+
+	gboolean all_white_input;
 } PlugInVals;
 
 typedef struct {
-	gboolean dummy;
+	gboolean preview;
 } PlugInUIVals;
 
 /*  Default values  */
