@@ -8,8 +8,8 @@
  */
 #include "config.h"
 
+#include <string.h>
 #include <gtk/gtk.h>
-
 #include <libgimp/gimp.h>
 
 #include "main.h"
@@ -18,103 +18,153 @@
 
 #include "plugin-intl.h"
 
-
-static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable *mask, PlugInVals *opts, gint image_id);
-inline static gfloat wsum2sel(gfloat w, gfloat lt, gfloat ht, gboolean logthres);
-static void slopes(GimpDrawable *drawable, GimpDrawable *mask,
-                   gint left, gint top, gint right, gint bottom,
-                   gboolean halfsel, gboolean leastsq,
-                   gint window, gfloat *kernel, gfloat **pdx, gfloat **pdy);
-inline static void gray_pixel(GimpImageType type, guchar *pixel);
-//inline static void find_center(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel, guchar *pixel,
-//                               gfloat *pcenterx, gfloat *outcy, gfloat *outcz);
-inline static void old_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel, guchar *center,
-                             gfloat *dxpixel, gfloat *dypixel);
-//inline static void old_slope2(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel,
-//                              gfloat *pcenterx, gfloat *pcentery, gfloat *centerz, gfloat *dxpixel, gfloat *dypixel);
-inline static void least_squares_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel,
-                                       //gfloat *pcenterx, gfloat *pcentery, gfloat *centerz,
-                                       guchar *center, gfloat *dxpixel, gfloat *dypixel);
-inline static void flat_slope(GimpImageType type, gfloat *dxpixel, gfloat *dypixel);
-static gfloat *make_kernel(InterpolateKernelType type, gint window, gfloat s);
-inline static gfloat gaussian(gfloat x, gfloat s);
+#define SKIP_UNSELECTED 1
 
 
-void render(gint32 image_ID, GimpDrawable *drawable, PlugInVals *vals) {
-	gimp_image_undo_group_start(image_ID);
-
-
-	/*
-	 * get input mask (first channel found, or selection)
-	 */
-	gint mask_id = 0;
-	GimpDrawable *mask = 0;
-
-	if (vals->usesel) {
-		gimp_selection_invert(image_ID);
-		mask_id = gimp_selection_save(image_ID);
-		gimp_selection_invert(image_ID);
-		mask = gimp_drawable_get(mask_id);
-	} else {
-		mask = gimp_drawable_get(vals->chanid);
-	}
-
-
-	/*
-	 * save selection to channel, so we can modify it
-	 */
-	gint sel_id = gimp_selection_save(image_ID);
-	GimpDrawable *sel = gimp_drawable_get(sel_id);
-
-
-	/*
-	 * interpolate
-	 */
-	interpolate(drawable, sel, mask, vals, image_ID);
-
-
-	/*
-	 * no more need for temporary selection channel
-	 */
-	gimp_drawable_detach(sel);
-	if (vals->changesel) {
-		gimp_selection_load(sel_id);
-	}
-	gimp_image_remove_channel(image_ID, sel_id);
-
-
-	/*
-	 * remove input mask channel if it is our temporary
-	 */
-	gimp_drawable_detach(mask);
-	if (mask_id) {
-		gimp_image_remove_channel(image_ID, mask_id);
-	}
-
-
-	gimp_image_undo_group_end(image_ID);
+typedef struct InterpolateKernel {
+	gfloat *values;
+	gint width, height;
+} InterpolateKernel;
+static void kernel_init(InterpolateKernel *kernel, InterpolateKernelType type, gint window, gfloat s);
+static void kernel_fini(InterpolateKernel *kernel);
+inline static gfloat kernel_value(InterpolateKernel *kernel, gint deltax, gint deltay) {
+	return kernel->values[kernel->width*(kernel->height/2+deltay) + (kernel->width/2+deltax)];
 }
 
 
-static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable *mask, PlugInVals *opts, gint image_id) {
-	gint x,y;
-	gint deltay, deltax;
+static void interpolate(GimpDrawable *drawable, PlugInVals *opts, gint image_id, GimpPreview *preview);
+inline static gfloat wsum2sel(gfloat w, gfloat lt, gfloat ht, gboolean logthres);
+static void slopes(GimpDrawable *drawable, GimpDrawable *mask,
+                   gint left, gint top, gint right, gint bottom,
+                   InterpolateKernel *kernel, gboolean halfsel, gboolean leastsq,
+                   gfloat **pdx, gfloat **pdy, GimpPreview *preview);
+inline static void old_slope(Ring *ring, Ring *mring, gint x,
+                             InterpolateKernel *kernel, gboolean halfsel,
+                             guchar *center, gfloat *dxpixel, gfloat *dypixel);
+inline static void least_squares_slope(Ring *ring, Ring *mring, gint x,
+                                       InterpolateKernel *kernel, gboolean halfsel,
+                                       guchar *center, gfloat *dxpixel, gfloat *dypixel);
+inline static void flat_slope(GimpImageType type, gfloat *dxpixel, gfloat *dypixel);
+
+
+inline static gfloat gaussian(gfloat x, gfloat s);
+inline static void gray_pixel(GimpImageType type, guchar *pixel);
+
+
+void invoke_render(InterpolateInvocation *inv, GimpPreview *preview) {
+	return render(inv->drawable, inv->vals, preview);
+}
+
+void render(GimpDrawable *drawable, PlugInVals *vals, GimpPreview *preview) {
+	gint32 image_id = gimp_drawable_get_image(drawable->drawable_id);
+
+	if (preview) {
+		gimp_image_undo_freeze(image_id);
+	} else {
+		gimp_image_undo_group_start(image_id);
+	}
+
+	interpolate(drawable, vals, image_id, preview);
+
+	if (preview) {
+		gimp_image_undo_thaw(image_id);
+	} else {
+		gimp_image_undo_group_end(image_id);
+	}
+}
+
+
+static void interpolate(GimpDrawable *drawable, PlugInVals *opts, gint image_id, GimpPreview *preview) {
+	GimpImageType drawable_type = gimp_drawable_type(drawable->drawable_id);
+	gint channels = gimp_drawable_bpp(drawable->drawable_id);
+
+	gboolean changesel = opts->changesel && !preview;
+	gboolean newlayer = opts->newlayer && !preview;
+	gboolean linnewlayer = opts->linnewlayer && !preview;
 
-	gint window = opts->window;
 	gboolean wslopes = opts->slopes;
 
 
 	gint left,top,right,bottom,width,height;
-	gimp_drawable_mask_bounds(drawable->drawable_id, &left, &top, &right, &bottom);
-	width = right-left;
-	height = bottom-top;
+	if (preview) {
+		gimp_preview_get_position(preview, &left, &top);
+		gimp_preview_get_size(preview, &width, &height);
+		right = left + width;
+		bottom = top + height;
+	} else {
+		gimp_drawable_mask_bounds(drawable->drawable_id, &left, &top, &right, &bottom);
+		width = right-left;
+		height = bottom-top;
+	}
+
+
+	/*
+	 * make input mask
+	 */
+	gint tmp_mask_id = -1;
+	GimpDrawable *mask = 0;
+
+	if (!opts->all_white_input && gimp_drawable_is_valid(opts->chanid)) {
+		if (opts->usesel) {
+			tmp_mask_id = gimp_channel_copy(opts->chanid);
+			gimp_image_add_channel(image_id, tmp_mask_id, 0);
+			mask = gimp_drawable_get(tmp_mask_id);
+		} else {
+			mask = gimp_drawable_get(opts->chanid);
+		}
+	}
+	
+	if (!mask) {
+		GimpRGB rgb;
+		gimp_rgb_set(&rgb, 0, 0, 0);
+		tmp_mask_id = gimp_channel_new(
+			image_id, "Input mask temporary",
+			gimp_image_width(image_id),
+			gimp_image_height(image_id),
+			0.5, &rgb
+		);
+		gimp_drawable_fill(tmp_mask_id, GIMP_WHITE_FILL);
+		gimp_image_add_channel(image_id, tmp_mask_id, 0);
+
+		mask = gimp_drawable_get(tmp_mask_id);
+	}
+
+	if (opts->usesel) {
+		GimpPixelRgn outm;
+		gimp_pixel_rgn_init(&outm, mask, left, top, width, height, TRUE, TRUE);
+		guchar *outmrow = g_new(guchar, mask->bpp*width);
+		memset(outmrow, 0, mask->bpp*width);
+		gint y;
+		for (y = 0 ; y < height ; ++y) {
+			gimp_pixel_rgn_set_row(&outm,outmrow,0,y,width);
+		}
+		gimp_drawable_flush(mask);
+		gimp_drawable_merge_shadow(tmp_mask_id, TRUE);
+	}
 
-	GimpImageType drawable_type = gimp_drawable_type(drawable->drawable_id);
-	gint channels = gimp_drawable_bpp(drawable->drawable_id);
 	gint mchannels = gimp_drawable_bpp(mask->drawable_id);
 
+
+	/*
+	 * save selection to channel, so we can modify it (or to skip unselected pixels)
+	 */
+	gint sel_id = -1;
+	GimpDrawable *sel = 0;
+	#ifdef SKIP_UNSELECTED
+		sel_id = gimp_selection_save(image_id);
+		sel = gimp_drawable_get(sel_id);
+	#else
+		if (changesel) {
+			sel_id = gimp_selection_save(image_id);
+			sel = gimp_drawable_get(sel_id);
+		}
+	#endif
+
+
+	/*
+	 * make the new layers, if we are making new layers
+	 */
 	GimpDrawable *outdrawable, *outslopesdrawable;
-	gboolean newlayer = opts->newlayer;
 	if (newlayer) {
 		gint layer_id = gimp_layer_new(
 			image_id,
@@ -137,7 +187,6 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 	gint outchannels = gimp_drawable_bpp(outdrawable->drawable_id);
 	gint outslopeschannels;
 
-	gboolean linnewlayer = opts->linnewlayer;
 	if (linnewlayer) {
 		gint layer_id = gimp_layer_new(
 			image_id,
@@ -159,23 +208,29 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 	}
 
 
-	gint inleft = MAX(left-window,0);
-	gint intop = MAX(top-window,0);
-	gint inwidth = MIN(right+window,drawable->width)-inleft;
-	gint inheight = MIN(bottom+window,drawable->height)-intop;
+	/*
+	 * make kernel
+	 */
+	InterpolateKernel kernel;
+	kernel_init(&kernel, opts->kernel, opts->window, opts->sigma);
 
 
 	/*
-	 * calculate slopes of sample areas
+	 * calculate slopes of sample areas (this is the first pass)
 	 */
+	gint inleft = MAX(left-kernel.width/2,0);
+	gint intop = MAX(top-kernel.height/2,0);
+	gint inwidth = MIN(right+kernel.width/2,drawable->width)-inleft;
+	gint inheight = MIN(bottom+kernel.height/2,drawable->height)-intop;
 	gfloat *dx=0, *dy=0;
 	if (wslopes) {
-		gfloat *linkernel = make_kernel(opts->linkernel,opts->linwindow,opts->linsigma);
+		InterpolateKernel linkernel;
+		kernel_init(&linkernel, opts->linkernel, opts->linwindow, opts->linsigma);
 		slopes(drawable, mask,
 		       inleft, intop, inleft+inwidth, intop+inheight,
-		       opts->halfsel, opts->leastsq,
-		       opts->linwindow, linkernel, &dx, &dy);
-		g_free(linkernel);
+		       &linkernel, opts->halfsel, opts->leastsq,
+		       &dx, &dy, preview);
+		kernel_fini(&linkernel);
 	}
 
 	gint tw = gimp_tile_width();
@@ -183,81 +238,84 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 
 
 	/*
-	 * make kernel layer (brush?)
+	 * get input/output pixel regions for interpolation pass
 	 */
-	gint kernel_width = (1+2*window);
-	gfloat *kernel = make_kernel(opts->kernel,window,opts->sigma);
+	Ring in; ring_init(&in, drawable, left, top, right, bottom, kernel.width/2, kernel.height/2);
+	Ring inm; ring_init(&inm, mask, left, top, right, bottom, kernel.width/2, kernel.height/2);
 
-
-	/*
-	 * get input pixel regions; fill rings
-	 */
-	GimpPixelRgn in;
-	gimp_pixel_rgn_init(&in, drawable, inleft, intop, inwidth, inheight, FALSE, FALSE);
-
-	GimpPixelRgn min;
-	gimp_pixel_rgn_init(&min, mask, inleft, intop, inwidth, inheight, FALSE, FALSE);
-
-	gint in_ring = window;
-	gint inrow_bytes = channels*inwidth;
-	guchar *inrows = g_new(guchar, kernel_width*inrow_bytes);
-	gint inmrow_bytes = mchannels*inwidth;
-	guchar *inmrows = g_new(guchar, kernel_width*inmrow_bytes);
-	for (y = MAX(top-window,0) ;
-	     y <= MIN(top+window,drawable->height-1) ; ++y) {
-		gimp_pixel_rgn_get_row(&in,
-			inrows + (in_ring+y-top)*inrow_bytes,
-			inleft, y, inwidth
-		);
-		gimp_pixel_rgn_get_row(&min,
-			inmrows + (in_ring+y-top)*inmrow_bytes,
-			inleft, y, inwidth
-		);
-	}
-
-	GimpPixelRgn ins;
-	gimp_pixel_rgn_init(&ins, sel, left, top, width, height, FALSE, FALSE);
-	guchar *insrow = g_new(guchar, mchannels*width);
-
-
-	/*
-	 * get output pixel regions
-	 */
 	GimpPixelRgn out;
 	if (newlayer) {
 		gimp_pixel_rgn_init(&out, outdrawable, 0, 0, width, height, TRUE, TRUE);
 	} else {
-		gimp_pixel_rgn_init(&out, outdrawable, left, top, width, height, TRUE, TRUE);
+		gimp_pixel_rgn_init(&out, outdrawable, left, top, width, height, !preview, TRUE);
 	}
 	guchar *outrow = g_new(guchar, outchannels*width);
 
 	GimpPixelRgn outslopes;
+	guchar *outslopesrow = 0;
 	if (linnewlayer) {
 		gimp_pixel_rgn_init(&outslopes, outslopesdrawable, 0, 0, width, height, TRUE, TRUE);
+		outslopesrow = g_new(guchar, outslopeschannels*width);
 	}
-	guchar *outslopesrow = g_new(guchar, outslopeschannels*width);
+
+
+	/*
+	 * get selection, to skip unselected pixels
+	 */
+	GimpPixelRgn ins;
+	guchar *insrow = 0;
+	#ifdef SKIP_UNSELECTED
+		gimp_pixel_rgn_init(&ins, sel, left, top, width, height, FALSE, FALSE);
+		insrow = g_new(guchar, mchannels*width);
+	#else
+		if (changesel) {
+			gimp_pixel_rgn_init(&ins, sel, left, top, width, height, FALSE, FALSE);
+			insrow = g_new(guchar, mchannels*width);
+		}
+	#endif
 
 	GimpPixelRgn outs;
-	if (opts->changesel) {
+	guchar *outsrow = 0;
+	if (changesel) {
 		gimp_pixel_rgn_init(&outs, sel, left, top, width, height, TRUE, TRUE);
+		outsrow = g_new(guchar, mchannels*width);
 	}
-	guchar *outsrow = g_new(guchar, mchannels*width);
 
 
 	/*
 	 * do the interpolation
 	 */
-	gimp_progress_init(_("Interpolating..."));
-	for (y = top ; y < bottom ; ++y) {
-		guchar *inmrow = inmrows + ( (in_ring%kernel_width)*inmrow_bytes );
-		gimp_pixel_rgn_get_row(&ins, insrow, left, y, width);
+	gint x;
+	gint deltay, deltax;
+	if (!preview) gimp_progress_init(_("Interpolating..."));
+	for (; ring_loop(&in) && ring_loop(&inm) ; ring_rotate(&in), ring_rotate(&inm)) {
+		guchar *inmrow = 0;
+		if (changesel) {
+			inmrow = ring_row(&inm);
+		}
+
+		#ifdef SKIP_UNSELECTED
+			gimp_pixel_rgn_get_row(&ins, insrow, left, in.y, width);
+		#else
+			if (changesel) {
+				gimp_pixel_rgn_get_row(&ins, insrow, left, in.y, width);
+			}
+		#endif
 
 		for (x = left ; x < right ; ++x) {
-			guchar *inmpixel = inmrow + (x-inleft)*mchannels;
-			guchar *inspixel = insrow + (x-left)*mchannels;
+			guchar *inspixel = 0;
+			#ifdef SKIP_UNSELECTED
+				inspixel = insrow + (x-left)*mchannels;
+			#else
+				if (changesel) {
+					inspixel = insrow + (x-left)*mchannels;
+				}
+			#endif
 			guchar *outpixel = outrow + (x-left)*outchannels;
-			guchar *outslopespixel = outslopesrow + (x-left)*outslopeschannels;
-			guchar *outspixel = outsrow + (x-left)*mchannels;  // opts->changesel?
+			guchar *outslopespixel = 0;
+			if (linnewlayer) {
+				outslopespixel = outslopesrow + (x-left)*outslopeschannels;
+			}
 
 			gfloat sum[4];
 			sum[0] = 0.0;
@@ -271,56 +329,35 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 			dsum[3] = 0.0;
 			gfloat ws = 0.0;
 
-			gint deltay_min = MAX(-window,-y);
-			gint deltay_max = MIN(window,drawable->height-1-y);
+			gint deltay_min = ring_min_deltay(&in);
+			gint deltay_max = ring_max_deltay(&in);
+			#ifdef SKIP_UNSELECTED
+			if (inspixel[0])
+			#endif
 			for (deltay = deltay_min ; deltay <= deltay_max ; ++deltay) {
-				guchar *inrow2 = inrows + ( ((in_ring+deltay)%kernel_width)*inrow_bytes );
-				guchar *inmrow2 = inmrows + ( ((in_ring+deltay)%kernel_width)*inmrow_bytes );
+				guchar *inrow2 = ring_row_delta(&in, deltay);
+				guchar *inmrow2 = ring_row_delta(&inm, deltay);
+
 				gfloat *dxrow = 0;
 				gfloat *dyrow = 0;
-
-				gint deltax_min = MAX(-window,-x);
-				gint deltax_max = MIN(window,drawable->width-1-x);
-
 				if (wslopes) {
-					dxrow = dx + channels*inwidth*(y+deltay-intop);
-					dyrow = dy + channels*inwidth*(y+deltay-intop);
+					dxrow = dx + channels*inwidth*(in.y+deltay-intop);
+					dyrow = dy + channels*inwidth*(in.y+deltay-intop);
 				}
 
+				gint deltax_min = ring_min_deltax(&in, x);
+				gint deltax_max = ring_max_deltax(&in, x);
 				for (deltax = deltax_min ; deltax <= deltax_max ; ++deltax) {
-					guchar *inpixel2 = inrow2 + (x+deltax-inleft)*channels;
-					guchar *inmpixel2 = inmrow2 + (x+deltax-inleft)*mchannels;
+					guchar *inpixel2 = inrow2 + (x+deltax-left)*channels;
+					guchar *inmpixel2 = inmrow2 + (x+deltax-left)*mchannels;
 
-					gfloat w = kernel[kernel_width*(window+deltay) + window+deltax];
+					gfloat w = kernel_value(&kernel, deltax, deltay);
 
 					if ( ! ( opts->halfsel ? (inmpixel2[0]>0) : (inmpixel2[0]==255) ) ) {
 						continue;
 					}
 
-					if (w<0.0) {
-						/*
-						 * -1 means infinity. this input pixel has infinite weight
-						 * (center pixel in 'inverse' kernels)
-						 */
-						ws = opts->highthres;
-						switch (drawable_type) {
-						case GIMP_RGBA_IMAGE:
-							sum[3] = opts->highthres*inpixel2[3];
-						case GIMP_RGB_IMAGE:
-							sum[2] = opts->highthres*inpixel2[2];
-						case GIMP_GRAYA_IMAGE:
-							sum[1] = opts->highthres*inpixel2[1];
-						case GIMP_GRAY_IMAGE: default:
-							sum[0] = opts->highthres*inpixel2[0];
-							break;
-						}
-						dsum[0] = 0.0;
-						dsum[1] = 0.0;
-						dsum[2] = 0.0;
-						dsum[3] = 0.0;
-						goto breakout;
-
-					} else {
+					if (w>=0.0) {
 						/*
 						 * ordinary input pixel
 						 */
@@ -357,6 +394,28 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 						}
 
 						ws += w;
+					} else {
+						/*
+						 * -1 means infinity. this input pixel has infinite weight
+						 * (center pixel in 'inverse' kernels)
+						 */
+						ws = opts->highthres;
+						switch (drawable_type) {
+						case GIMP_RGBA_IMAGE:
+							sum[3] = opts->highthres*inpixel2[3];
+						case GIMP_RGB_IMAGE:
+							sum[2] = opts->highthres*inpixel2[2];
+						case GIMP_GRAYA_IMAGE:
+							sum[1] = opts->highthres*inpixel2[1];
+						case GIMP_GRAY_IMAGE: default:
+							sum[0] = opts->highthres*inpixel2[0];
+							break;
+						}
+						dsum[0] = 0.0;
+						dsum[1] = 0.0;
+						dsum[2] = 0.0;
+						dsum[3] = 0.0;
+						goto breakout;
 					}
 				}
 			}
@@ -368,7 +427,6 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 			 */
 			if (wslopes) {
 				if (linnewlayer) {
-
 					if (ws>0.0) {
 						switch (drawable_type) {
 						case GIMP_GRAY_IMAGE: default:
@@ -438,8 +496,13 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 				gray_pixel(outdrawable_type, outpixel);
 			}
 
-			if (opts->changesel) {
-				// input mask is multiplied with pixels without being divided by 255, so divide here
+			if (changesel) {
+				guchar *inmpixel = inmrow + (x-left)*mchannels;
+				guchar *outspixel = outsrow + (x-left)*mchannels;
+
+				/*
+				 * input mask is multiplied with pixels without being divided by 255, so divide here
+				 */
 				if (opts->halfsel) ws /= 255.0;
 
 				gfloat outsel = wsum2sel(ws, opts->lowthres, opts->highthres, opts->logthres) * inspixel[0];
@@ -449,80 +512,72 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 		}
 
 		if (newlayer) {
-			gimp_pixel_rgn_set_row(&out,outrow,0,y-top,right-left);
+			gimp_pixel_rgn_set_row(&out,outrow,0,in.y-top,right-left);
 		} else {
-			gimp_pixel_rgn_set_row(&out,outrow,left,y,right-left);
+			gimp_pixel_rgn_set_row(&out,outrow,left,in.y,right-left);
 		}
 		if (linnewlayer) {
-			gimp_pixel_rgn_set_row(&outslopes,outslopesrow,0,y-top,right-left);
+			gimp_pixel_rgn_set_row(&outslopes,outslopesrow,0,in.y-top,right-left);
 		}
-		if (opts->changesel) {
-			gimp_pixel_rgn_set_row(&outs,outsrow,left,y,right-left);
+		if (changesel) {
+			gimp_pixel_rgn_set_row(&outs,outsrow,left,in.y,right-left);
 		}
 
-
-		/*
-		 * rotate input row buffers, to move to the next row
-		 */
-		++in_ring;
-		if ((y+1)+window < drawable->height) {
-			gimp_pixel_rgn_get_row(&in,
-				inrows + ((in_ring+window)%kernel_width)*inrow_bytes,
-				inleft, (y+1)+window, inwidth
-			);
-			gimp_pixel_rgn_get_row(&min,
-				inmrows + ((in_ring+window)%kernel_width)*inmrow_bytes,
-				inleft, (y+1)+window, inwidth
-			);
-		}
-
-		if (!(y%4)) {
-			gimp_progress_update(1.0*(y-top+1)/(bottom-top));
+		if (!preview && !(in.y%4)) {
+			gimp_progress_update(1.0*(in.y-top+1)/(bottom-top));
 		}
 	}
-	gimp_progress_end();
+	if (!preview) gimp_progress_end();
 
 
 	/*
 	 * done with the pixel regions, kernel and slopes
 	 */
-	g_free(outsrow);
-	g_free(outslopesrow);
+	if (changesel) {
+		g_free(outsrow);
+	}
+	if (linnewlayer) {
+		g_free(outslopesrow);
+	}
 	g_free(outrow);
 
-	g_free(inmrows);
-	g_free(inrows);
+	ring_fini(&inm);
+	ring_fini(&in);
 
-	g_free(kernel);
+	kernel_fini(&kernel);
 
 	g_free(dx);
 	g_free(dy);
 
 
 	/*
-	 * we can just flush here
+	 * the plumbing code
 	 */
-	gimp_drawable_flush(outdrawable);
-	gimp_drawable_merge_shadow(outdrawable->drawable_id, TRUE);
-	if (newlayer) {
-		gimp_drawable_update(outdrawable->drawable_id, 0, 0, width, height);
-		gimp_drawable_detach(outdrawable);
+	if (preview) {
+		gimp_drawable_preview_draw_region(GIMP_DRAWABLE_PREVIEW(preview), &out);
 	} else {
-		gimp_drawable_update(outdrawable->drawable_id, left, top, width, height);
-	}
+		gimp_drawable_flush(outdrawable);
+		gimp_drawable_merge_shadow(outdrawable->drawable_id, TRUE);
+		if (newlayer) {
+			gimp_drawable_update(outdrawable->drawable_id, 0, 0, width, height);
+			gimp_drawable_detach(outdrawable);
+		} else {
+			gimp_drawable_update(outdrawable->drawable_id, left, top, width, height);
+		}
 
-	if (linnewlayer) {
-		gimp_drawable_flush(outslopesdrawable);
-		gimp_drawable_merge_shadow(outslopesdrawable->drawable_id, TRUE);
-		gimp_drawable_update(outslopesdrawable->drawable_id, 0, 0, width, height);
-		gimp_drawable_detach(outslopesdrawable);
+		if (linnewlayer) {
+			gimp_drawable_flush(outslopesdrawable);
+			gimp_drawable_merge_shadow(outslopesdrawable->drawable_id, TRUE);
+			gimp_drawable_update(outslopesdrawable->drawable_id, 0, 0, width, height);
+			gimp_drawable_detach(outslopesdrawable);
+		}
 	}
 
 
 	/*
 	 * only remove the selection if we are modifying it
 	 */
-	if (opts->changesel) {
+	if (changesel) {
 		gint tmp_sel_id = gimp_selection_save(image_id);
 		gimp_selection_none(image_id);
 
@@ -533,6 +588,31 @@ static void interpolate(GimpDrawable *drawable, GimpDrawable *sel, GimpDrawable
 		gimp_selection_load(tmp_sel_id);
 		gimp_image_remove_channel(image_id, tmp_sel_id);
 	}
+
+
+	/*
+	 * no more need for temporary selection channel
+	 */
+	if (changesel) {
+		gimp_selection_load(sel_id);
+		#ifndef SKIP_UNSELECTED
+			gimp_drawable_detach(sel);
+			gimp_image_remove_channel(image_id, sel_id);
+		#endif
+	}
+	#ifdef SKIP_UNSELECTED
+		gimp_drawable_detach(sel);
+		gimp_image_remove_channel(image_id, sel_id);
+	#endif
+
+
+	/*
+	 * remove input mask channel if it is our temporary
+	 */
+	gimp_drawable_detach(mask);
+	if (tmp_mask_id != -1) {
+		gimp_image_remove_channel(image_id, tmp_mask_id);
+	}
 }
 
 
@@ -553,160 +633,54 @@ inline static gfloat wsum2sel(gfloat w, gfloat lt, gfloat ht, gboolean logthres)
 
 static void slopes(GimpDrawable *drawable, GimpDrawable *mask,
                    gint left, gint top, gint right, gint bottom,
-                   gboolean halfsel, gboolean leastsq,
-                   gint window, gfloat *kernel, gfloat **pdx, gfloat **pdy) {
+                   InterpolateKernel *kernel, gboolean halfsel, gboolean leastsq,
+                   gfloat **pdx, gfloat **pdy, GimpPreview *preview) {
 	gint width = right-left;
 	gint height = bottom-top;
 
-	Ring ring;
-	ring_init(&ring, drawable, left, top, right, bottom, window);
-
-	Ring mring;
-	ring_init(&mring, mask, left, top, right, bottom, window);
-
+	Ring ring; ring_init(&ring, drawable, left, top, right, bottom, kernel->width/2, kernel->height/2);
+	Ring mring; ring_init(&mring, mask, left, top, right, bottom, kernel->width/2, kernel->height/2);
 	gfloat *dx = *pdx = g_new(gfloat, width*height*ring.bpp);
 	gfloat *dy = *pdy = g_new(gfloat, width*height*ring.bpp);
 
-	gimp_progress_init(_("Calculating slopes..."));
+	if (!preview) gimp_progress_init(_("Calculating slopes..."));
 	for (; ring_loop(&ring) && ring_loop(&mring) ; ring_rotate(&ring), ring_rotate(&mring)) {
-		guchar *row = ring_row(&ring,ring.y);
-		guchar *mrow = ring_row(&mring,mring.y);
+		guchar *row = ring_row(&ring);
+		guchar *mrow = ring_row(&mring);
 		gfloat *dxrow = dx + ring.bpp*width*(ring.y-top);
 		gfloat *dyrow = dy + ring.bpp*width*(ring.y-top);
 
 		gint x;
 		for (x = left ; x < right ; ++x) {
-			guchar *pixel = row + ( (x-left)*ring.bpp );  // center pixel
+			guchar *pixel = row + ( (x-left)*ring.bpp );  /* center pixel */
 			guchar *mpixel = mrow + ( (x-left)*mring.bpp );
 			gfloat *dxpixel = dxrow + ring.bpp*(x-left);
 			gfloat *dypixel = dyrow + ring.bpp*(x-left);
 
+			/*
+			 * calculate average of nearby pixels
+			 */
 			if (halfsel ? !mpixel[0] : mpixel[0]<255) {
 				flat_slope(ring.type, dxpixel, dypixel);
 			} else if (leastsq) {
-				//gfloat centerx;
-				//gfloat centery;
-				//gfloat centerz[4];
-				//find_center(&ring, &mring, x, kernel, window, halfsel, pixel, &centerx, &centery, centerz);
-				least_squares_slope(&ring, &mring, x, kernel, window, halfsel, pixel, dxpixel, dypixel);
+				least_squares_slope(&ring, &mring, x, kernel, halfsel, pixel, dxpixel, dypixel);
 			} else {
-				old_slope(&ring, &mring, x, kernel, window, halfsel, pixel, dxpixel, dypixel);
+				old_slope(&ring, &mring, x, kernel, halfsel, pixel, dxpixel, dypixel);
 			}
 		}
 
-		if (!(ring.y%4)) {
+		if (!preview && !(ring.y%4)) {
 			gimp_progress_update(1.0*(ring.y-top+1)/(bottom-top));
 		}
 	}
-	gimp_progress_end();
+	if (!preview) gimp_progress_end();
 	ring_fini(&ring);
 }
 
 
-inline static void gray_pixel(GimpImageType type, guchar *pixel) {
-	switch (type) {
-	case GIMP_RGBA_IMAGE:
-		pixel[3] = 128;
-	case GIMP_RGB_IMAGE:
-		pixel[2] = 128;
-	case GIMP_GRAYA_IMAGE:
-		pixel[1] = 128;
-	case GIMP_GRAY_IMAGE: default:
-		pixel[0] = 128;
-		break;
-	}
-}
-
-
-#if 0
-inline static void find_center(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel, guchar *pixel,
-                               gfloat *pcenterx, gfloat *pcentery, gfloat *centerz) {
-	gint y = ring->y;
-	gfloat wsum = 0.0;
-	gfloat xsum = 0.0;
-	gfloat ysum = 0.0;
-	gfloat zsum[4];
-	zsum[0] = 0.0;
-	zsum[1] = 0.0;
-	zsum[2] = 0.0;
-	zsum[3] = 0.0;
-
-	gint y2_min = MAX(0,y-window);
-	gint y2_max = MIN(y+window,ring->dr->height-1);
-	gint y2;
-	for (y2 = y2_min ; y2 <= y2_max ; ++y2) {
-		guchar *row = ring_row(ring,y2);
-		guchar *mrow = ring_row(mring,y2);
-
-		gint x2_min = MAX(0,x-window);
-		gint x2_max = MIN(x+window,ring->dr->width-1);
-		gint x2;
-		for (x2 = x2_min ; x2 <= x2_max ; ++x2) {
-			guchar *mpixel = mrow + ( (x2-ring->left)*mring->bpp );
-			if (halfsel ? mpixel[0]>0 : mpixel[0]==255) {
-				gfloat w = kernel[(1+2*window)*(window+y2-y) + (window+x2-x)];
-				if (halfsel) {
-					w *= mpixel[0];
-				}
-				if (w>0) {
-					wsum += w;
-					xsum += w*(x2-x);
-					ysum += w*(y2-y);
-
-					guchar *pixel = row + ( (x2-ring->left)*ring->bpp );
-					switch (ring->type) {
-					case GIMP_RGBA_IMAGE:
-						zsum[3] = w*pixel[3];
-					case GIMP_RGB_IMAGE:
-						zsum[2] = w*pixel[2];
-					case GIMP_GRAYA_IMAGE:
-						zsum[1] = w*pixel[1];
-					case GIMP_GRAY_IMAGE: default:
-						zsum[0] = w*pixel[0];
-						break;
-					}
-				}
-			}
-		}
-	}
-
-	if (wsum>0) {
-		*pcenterx = xsum/wsum;
-		*pcentery = ysum/wsum;
-		switch (ring->type) {
-		case GIMP_RGBA_IMAGE:
-			centerz[3] = zsum[3]/wsum;
-		case GIMP_RGB_IMAGE:
-			centerz[2] = zsum[2]/wsum;
-		case GIMP_GRAYA_IMAGE:
-			centerz[1] = zsum[1]/wsum;
-		case GIMP_GRAY_IMAGE: default:
-			centerz[0] = zsum[0]/wsum;
-			break;
-		}
-	} else {
-		*pcenterx = 0;
-		*pcentery = 0;
-		switch (ring->type) {
-		case GIMP_RGBA_IMAGE:
-			centerz[3] = pixel[3];
-		case GIMP_RGB_IMAGE:
-			centerz[2] = pixel[2];
-		case GIMP_GRAYA_IMAGE:
-			centerz[1] = pixel[1];
-		case GIMP_GRAY_IMAGE: default:
-			centerz[0] = pixel[0];
-			break;
-		}
-	}
-}
-#endif
-
-
-inline static void old_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel, guchar *center,
-                             gfloat *dxpixel, gfloat *dypixel) {
-	gint y = ring->y;
-
+inline static void old_slope(Ring *ring, Ring *mring, gint x,
+                             InterpolateKernel *kernel, gboolean halfsel,
+                             guchar *center, gfloat *dxpixel, gfloat *dypixel) {
 	gfloat wsum = 0.0;
 	gfloat dxsum[4];
 	dxsum[0] = 0.0;
@@ -719,28 +693,28 @@ inline static void old_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gi
 	dysum[2] = 0.0;
 	dysum[3] = 0.0;
 
-	gint y2_min = MAX(0,y-window);
-	gint y2_max = MIN(y+window,ring->dr->height-1);
-	gint y2;
-	for (y2 = y2_min ; y2 <= y2_max ; ++y2) {
-		guchar *row = ring_row(ring,y2);
-		guchar *mrow = ring_row(mring,y2);
+	gint deltay_min = ring_min_deltay(ring);
+	gint deltay_max = ring_max_deltay(ring);
+	gint deltay;
 
-		gint x2_min = MAX(0,x-window);
-		gint x2_max = MIN(x+window,ring->dr->width-1);
-		gint x2;
-		for (x2 = x2_min ; x2 <= x2_max ; ++x2) {
-			guchar *mpixel = mrow + ( (x2-ring->left)*mring->bpp );
-			gfloat deltax = x2-x;
-			gfloat deltay = y2-y;
+	gint deltax_min = ring_min_deltax(ring, x);
+	gint deltax_max = ring_max_deltax(ring, x);
+	gint deltax;
+
+	for (deltay = deltay_min ; deltay <= deltay_max ; ++deltay) {
+		guchar *row = ring_row_delta(ring, deltay);
+		guchar *mrow = ring_row_delta(mring, deltay);
+
+		for (deltax = deltax_min ; deltax <= deltax_max ; ++deltax) {
+			guchar *mpixel = mrow + ( (x+deltax-ring->left)*mring->bpp );
 			gfloat distsquared = (deltax*deltax + deltay*deltay);
 			if ( distsquared>0.0 && (halfsel ? mpixel[0]>0 : mpixel[0]==255) ) {
-				gfloat w = kernel[(1+2*window)*(window+y2-y) + (window+x2-x)];
+				gfloat w = kernel_value(kernel, deltax, deltay);
 				if (halfsel) {
 					w *= mpixel[0];
 				}
 				if (w>0) {
-					guchar *pixel = row + ( (x2-ring->left)*ring->bpp );
+					guchar *pixel = row + ( (x+deltax-ring->left)*ring->bpp );
 					switch (ring->type) {
 					case GIMP_RGBA_IMAGE:
 						dxsum[3] += w*(pixel[3]-center[3])*deltax/distsquared;
@@ -784,94 +758,9 @@ inline static void old_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gi
 }
 
 
-#if 0
-inline static void old_slope2(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel,
-                             gfloat *pcenterx, gfloat *pcentery, gfloat *centerz, gfloat *dxpixel, gfloat *dypixel) {
-	gint y = ring->y;
-
-	gfloat wsum = 0.0;
-	gfloat dxsum[4];
-	dxsum[0] = 0.0;
-	dxsum[1] = 0.0;
-	dxsum[2] = 0.0;
-	dxsum[3] = 0.0;
-	gfloat dysum[4];
-	dysum[0] = 0.0;
-	dysum[1] = 0.0;
-	dysum[2] = 0.0;
-	dysum[3] = 0.0;
-
-	gint y2_min = MAX(0,y-window);
-	gint y2_max = MIN(y+window,ring->dr->height-1);
-	gint y2;
-	for (y2 = y2_min ; y2 <= y2_max ; ++y2) {
-		guchar *row = ring_row(ring,y2);
-		guchar *mrow = ring_row(mring,y2);
-
-		gint x2_min = MAX(0,x-window);
-		gint x2_max = MIN(x+window,ring->dr->width-1);
-		gint x2;
-		for (x2 = x2_min ; x2 <= x2_max ; ++x2) {
-			guchar *mpixel = mrow + ( (x2-ring->left)*mring->bpp );
-			gfloat deltax = x2-x-*pcenterx;
-			gfloat deltay = y2-y-*pcentery;
-			gfloat distsquared = (deltax*deltax + deltay*deltay);
-			if ( distsquared>0.0 && (halfsel ? mpixel[0]>0 : mpixel[0]==255) ) {
-				gfloat w = kernel[(1+2*window)*(window+y2-y) + (window+x2-x)];
-				if (halfsel) {
-					w *= mpixel[0];
-				}
-				if (w>0) {
-					guchar *pixel = row + ( (x2-ring->left)*ring->bpp );
-					switch (ring->type) {
-					case GIMP_RGBA_IMAGE:
-						dxsum[3] += w*(pixel[3]-centerz[3])*deltax/distsquared;
-						dysum[3] += w*(pixel[3]-centerz[3])*deltay/distsquared;
-					case GIMP_RGB_IMAGE:
-						dxsum[2] += w*(pixel[2]-centerz[2])*deltax/distsquared;
-						dysum[2] += w*(pixel[2]-centerz[2])*deltay/distsquared;
-					case GIMP_GRAYA_IMAGE:
-						dxsum[1] += w*(pixel[1]-centerz[1])*deltax/distsquared;
-						dysum[1] += w*(pixel[1]-centerz[1])*deltay/distsquared;
-					case GIMP_GRAY_IMAGE: default:
-						dxsum[0] += w*(pixel[0]-centerz[0])*deltax/distsquared;
-						dysum[0] += w*(pixel[0]-centerz[0])*deltay/distsquared;
-						break;
-					}
-					wsum += w;
-				}
-			}
-		}
-	}
-
-	if (wsum>0.0) {
-		switch (ring->type) {
-		case GIMP_RGBA_IMAGE:
-			dxpixel[3] = 2*dxsum[3]/wsum;
-			dypixel[3] = 2*dysum[3]/wsum;
-		case GIMP_RGB_IMAGE:
-			dxpixel[2] = 2*dxsum[2]/wsum;
-			dypixel[2] = 2*dysum[2]/wsum;
-		case GIMP_GRAYA_IMAGE:
-			dxpixel[1] = 2*dxsum[1]/wsum;
-			dypixel[1] = 2*dysum[1]/wsum;
-		case GIMP_GRAY_IMAGE: default:
-			dxpixel[0] = 2*dxsum[0]/wsum;
-			dypixel[0] = 2*dysum[0]/wsum;
-			break;
-		}
-	} else {
-		flat_slope(ring->type, dxpixel, dypixel);
-	}
-}
-#endif
-
-
-inline static void least_squares_slope(Ring *ring, Ring *mring, gint x, gfloat *kernel, gint window, gboolean halfsel,
-                                       //gfloat *pcenterx, gfloat *pcentery, gfloat *centerz,
+inline static void least_squares_slope(Ring *ring, Ring *mring, gint x,
+                                       InterpolateKernel *kernel, gboolean halfsel,
                                        guchar *center, gfloat *dxpixel, gfloat *dypixel) {
-	gint y = ring->y;
-
 	gfloat ws = 0.0;
 
 	gfloat m[2][2];
@@ -890,29 +779,29 @@ inline static void least_squares_slope(Ring *ring, Ring *mring, gint x, gfloat *
 	v[3][0] = 0.0;
 	v[3][1] = 0.0;
 
-	gint y2_min = MAX(0,y-window);
-	gint y2_max = MIN(y+window,ring->dr->height-1);
-	gint y2;
-	for (y2 = y2_min ; y2 <= y2_max ; ++y2) {
-		guchar *row = ring_row(ring,y2);
-		guchar *mrow = ring_row(mring,y2);
+	gint deltay_min = ring_min_deltay(ring);
+	gint deltay_max = ring_max_deltay(ring);
+	gint deltay;
 
-		gint x2_min = MAX(0,x-window);
-		gint x2_max = MIN(x+window,ring->dr->width-1);
-		gint x2;
-		for (x2 = x2_min ; x2 <= x2_max ; ++x2) {
-			guchar *mpixel = mrow + ( (x2-ring->left)*mring->bpp );
-			gfloat deltax = x2-x;
-			gfloat deltay = y2-y;
+	gint deltax_min = ring_min_deltax(ring, x);
+	gint deltax_max = ring_max_deltax(ring, x);
+	gint deltax;
+
+	for (deltay = deltay_min ; deltay <= deltay_max ; ++deltay) {
+		guchar *row = ring_row_delta(ring, deltay);
+		guchar *mrow = ring_row_delta(mring, deltay);
+
+		for (deltax = deltax_min ; deltax <= deltax_max ; ++deltax) {
+			guchar *mpixel = mrow + ( (x+deltax-ring->left)*mring->bpp );
 			if ( halfsel ? mpixel[0]>0 : mpixel[0]==255 ) {
-				gfloat w = kernel[(1+2*window)*(window+y2-y) + (window+x2-x)];
+				gfloat w = kernel_value(kernel, deltax, deltay);
 				if (halfsel) {
 					w *= mpixel[0];
 				}
 				if (w>0) {
 					ws += w;
 
-					guchar *pixel = row + ( (x2-ring->left)*ring->bpp );
+					guchar *pixel = row + ( (x+deltax-ring->left)*ring->bpp );
 
 					m[0][0] += deltax*deltax*w;
 					m[0][1] += deltax*deltay*w;
@@ -1009,47 +898,95 @@ inline static void flat_slope(GimpImageType type, gfloat *dxpixel, gfloat *dypix
 }
 
 
-static gfloat *make_kernel(InterpolateKernelType type, gint window, gfloat s) {
-	gint kernel_width = (1+2*window);
-	gfloat *kernel = g_new(gfloat, kernel_width*kernel_width);
+static void kernel_init(InterpolateKernel *kernel, InterpolateKernelType type, gint window, gfloat s) {
+	gint bwidth=0, bheight=0;
+	gint mask_bpp=0, num_mask_bytes=0;
+	guint8 *mask_bytes=0;
+	gchar *brush_name = gimp_context_get_brush();
+	gint color_bpp=0, num_color_bytes=0;
+	guint8 *color_bytes=0;
+	if (type==KERNEL_BRUSH) {
+		gimp_brush_get_pixels(brush_name, &bwidth, &bheight,
+		                      &mask_bpp, &num_mask_bytes, &mask_bytes,
+		                      &color_bpp, &num_color_bytes, &color_bytes);
+		kernel->width = (bwidth/2)*2 + 1;
+		kernel->height = (bheight/2)*2 + 1;
+	} else {
+		kernel->width = kernel->height = 1 + 2*window;
+	}
+
+	kernel->values = g_new(gfloat, kernel->width*kernel->height);
+
 	gint y;
-	for (y = -window ; y <= window ; ++y) {
-		gfloat *frow = kernel + kernel_width*(window+y);
+	for (y = 0 ; y < kernel->height ; ++y) {
+		gint deltay = y - kernel->height/2;
+
+		gfloat *krow = kernel->values + kernel->width*y;
+
 		gint x;
-		for (x = -window ; x <= window ; ++x) {
-			gfloat d = sqrt(x*x+y*y);
+		for (x = 0 ; x < kernel->width ; ++x) {
+			gint deltax = x - kernel->width/2;
+
+			gfloat d = sqrt(deltax*deltax+deltay*deltay);
+
 			switch (type) {
 			case KERNEL_CONSTANT:
-				frow[window+x] = 1.0;
+				krow[x] = 1.0;
 				break;
 			case KERNEL_LINEAR:
-				frow[window+x] = MAX(0.0,1.0-d/s);
+				krow[x] = MAX(0.0,1.0-d/s);
 				break;
 			case KERNEL_SQUARE:
-				frow[window+x] = MAX(0.0,1.0-(d/s)*(d/s));
+				krow[x] = MAX(0.0,1.0-(d/s)*(d/s));
 				break;
 			case KERNEL_INVLINEAR:
-				frow[window+x] = (d>0.0 ? 1.0/d : -1.0);
+				krow[x] = (d>0.0 ? 1.0/d : -1.0);
 				break;
 			case KERNEL_INVSQUARE:
-				frow[window+x] = (d>0.0 ? 1.0/(d*d) : -1.0);
+				krow[x] = (d>0.0 ? 1.0/(d*d) : -1.0);
 				break;
 			case KERNEL_EXPONENTIAL:
-				frow[window+x] = exp(-d/s);
+				krow[x] = exp(-d/s);
 				break;
 			case KERNEL_GAUSSIAN:
-				frow[window+x] = gaussian(d,s);
+				krow[x] = gaussian(d,s);
 				break;
 			case KERNEL_INVNTH:
-				frow[window+x] = (d>0.0 ? pow(d,-s) : -1.0);
+				krow[x] = (d>0.0 ? pow(d,-s) : -1.0);
+				break;
+			case KERNEL_BRUSH:
+				if (x < bwidth && y < bheight) {
+					krow[x] = mask_bytes[mask_bpp*(bwidth*y + x)]/255.0;
+				} else {
+					krow[x] = 0.0;
+				}
 				break;
 			}
 		}
 	}
-	return kernel;
+}
+static void kernel_fini(InterpolateKernel *kernel) {
+	g_free(kernel->values);
 }
 
 
 inline static gfloat gaussian(gfloat x, gfloat s) {
 	return exp(-0.5*(x/s)*(x/s))/(s*sqrt(2*M_PI));
 }
+
+
+inline static void gray_pixel(GimpImageType type, guchar *pixel) {
+	switch (type) {
+	case GIMP_RGBA_IMAGE:
+		pixel[3] = 128;
+	case GIMP_RGB_IMAGE:
+		pixel[2] = 128;
+	case GIMP_GRAYA_IMAGE:
+		pixel[1] = 128;
+	case GIMP_GRAY_IMAGE: default:
+		pixel[0] = 128;
+		break;
+	}
+}
+
+
